class ReturnsError(Exception): ...


class OptionError(ReturnsError): ...


class OptionUnwrapError(OptionError): ...


class OptionExpectError(OptionError): ...


class ResultError(ReturnsError): ...


class ResultUnwrapError(ResultError): ...


class ResultExpectError(ResultError): ...
